/* usadelndsoc
 *
 * Copyright © 2022 Pauli Virtanen
 *    @author Pauli Virtanen <pauli.t.virtanen@jyu.fi>
 *
 * SPDX-License-Identifier: AGPL-3.0-or-later
 */
#ifndef COMMON_HPP_
#define COMMON_HPP_

#include <iostream>

#include <Eigen/Core>

#include <cppad/cppad.hpp>
#include <cppad/example/cppad_eigen.hpp>

#include <adept.h>

#include "array.hpp"
#include "vector.hpp"

typedef std::complex<double> Complex;
typedef CppAD::AD<Complex> ADComplex;

template <typename Scalar_, int Rows_, int Cols_>
using Matrix = Eigen::Matrix<Scalar_, Rows_, Cols_, Eigen::RowMajor>;

namespace Eigen
{
    /* Eigen binop maps: (double/complex, ADComplex) -> ADComplex */
    template<typename BinaryOp> struct ScalarBinaryOpTraits<ADComplex,int,BinaryOp> { typedef ADComplex ReturnType;  };
    template<typename BinaryOp> struct ScalarBinaryOpTraits<int,ADComplex,BinaryOp> { typedef ADComplex ReturnType;  };
    template<typename BinaryOp> struct ScalarBinaryOpTraits<ADComplex,double,BinaryOp> { typedef ADComplex ReturnType;  };
    template<typename BinaryOp> struct ScalarBinaryOpTraits<double,ADComplex,BinaryOp> { typedef ADComplex ReturnType;  };
    template<typename BinaryOp> struct ScalarBinaryOpTraits<ADComplex,Complex,BinaryOp> { typedef ADComplex ReturnType;  };
    template<typename BinaryOp> struct ScalarBinaryOpTraits<Complex,ADComplex,BinaryOp> { typedef ADComplex ReturnType;  };
}

namespace CppAD {
    ADComplex adcomplex(const ADComplex&& cpx) { return {cpx}; }
    ADComplex adcomplex(const Complex&& cpx) { return {cpx}; }

#define COMMON_HPP_DECLARE_BINOP(op) \
        ADComplex operator op(const ADComplex& a, const double& b) { return a op adcomplex(b); } \
        ADComplex operator op(const double& a, const ADComplex& b) { return adcomplex(a) op b; } \
        ADComplex operator op(const ADComplex& a, const Complex& b) { return a op adcomplex(b); } \
        ADComplex operator op(const Complex& a, const ADComplex& b) { return adcomplex(a) op b; }

    COMMON_HPP_DECLARE_BINOP(+);
    COMMON_HPP_DECLARE_BINOP(-);
    COMMON_HPP_DECLARE_BINOP(*);
    COMMON_HPP_DECLARE_BINOP(/);
#undef COMMON_HPP_DECLARE_BINOP
}

#endif
